@@ -12,8 +12,20 @@
  * Patches from * mb - Marcus Brinkmann <Marcus.Brinkmann@ruhr-uni-bochum.de>
  * for using get_current_dir_name() and portability on HURD. ps hacked this
  * so that it should (still?) compile on non-glibc systems.
+ * 2001-12-02 1.5.0
+ * Added autoconf support. Also added getopt.c, getopt1.c, getopt.h so that
+ * tmpreaper can be compiled on e.g. Solaris. It builds and runs on Solaris
+ * and netBSD in addition to Linux now.
+ * Ignore ext2fs immutable files.
+ * 2002-02-05 1.6.0
+ * ctime option added, check ctime in addition to atime, for those cases where
+ * files get created with an old date in atime and mtime (e.g. via Samba from
+ * a PC).
  */
 
+#define _LARGEFILE_SOURCE 1     /* handle >2GB files where possible */
+#define _FILE_OFFSET_BITS 64
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -22,6 +34,7 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <signal.h>
 #if HAVE_DIRENT_H
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
@@ -66,10 +79,11 @@
 # include <stdlib.h>
 #endif
 
+#include <fcntl.h>
+
 #ifdef __linux__
-# include <fcntl.h>
 # include <sys/ioctl.h>
-# include <linux/ext2_fs.h>
+# include <ext2fs/ext2_fs.h>
 # include <utime.h>
 #endif
 
@@ -125,6 +139,13 @@
 #define FLAGS_TEST		(1 << 6)
 #define FLAGS_TEST_P(fl)	((fl) & FLAGS_TEST)
 
+/* `--ctime': Use `ctime' in addition to `atime' */
+#define FLAGS_CTIME		(1 << 7)
+#define FLAGS_CTIME_P(fl)	((fl) & FLAGS_CTIME)
+
+/* The `--showdeleted' option was given */
+#define FLAGS_SHOWDEL           (1 << 8)
+#define FLAGS_SHOWDEL_P(fl)	((fl) & FLAGS_SHOWDEL)
 
 #define TIME_SECONDS	1
 #define TIME_MINUTES	60
@@ -142,7 +163,6 @@ typedef struct protect_entry
 
 protect_entry * protect_table;	/* Global */
 
-
 #define LOG_REALDEBUG	1
 #define LOG_DEBUG	2
 #define LOG_VERBOSE	3
@@ -151,6 +171,7 @@ protect_entry * protect_table;	/* Global */
 #define LOG_FATAL	6
 
 int	logLevel = LOG_NORMAL;
+int     delayMax = 0;
 
 void
 message (const int    level,
@@ -191,6 +212,16 @@ message (const int    level,
     }
 }
 
+void
+sig_alarmhandler(int sig)
+{
+    message(LOG_FATAL, "run time exceeded!\n"
+                       "This may be indicative of an attack to use tmpreaper to remove critical files;\n"
+                       "or the directories to clean up are excessive large and/or messed up.\n"
+                       "Please investigate.\n");
+    exit(1); /* redundant */
+}
+
 int
 safe_chdir (const char * dirname)
 {
@@ -465,6 +496,14 @@ cleanupDirectory (const char	   * dirname ,
 		else {
 		    if (sb.st_atime >= killTime)
 			continue;
+                    /* If ctime option is given, the ctime must also be so    */
+                    /* long ago. This is e.g. for samba as DOS preserves      */
+                    /* mtime when copying, so such files get removed too soon */
+                    /* Don't do it for directories though, as ctime may get   */
+                    /* reset below through utime()                            */
+                    if (!S_ISDIR(sb.st_mode) &&
+                        FLAGS_CTIME_P(flags) && sb.st_ctime >= killTime)
+                            continue;
 		}
 
 		if (S_ISDIR (sb.st_mode)) {
@@ -478,10 +517,15 @@ cleanupDirectory (const char	   * dirname ,
 			    message (LOG_VERBOSE,
 				     "Removing directory `%s'.\n",
 				     ent->d_name);
-			    if (rmdir (ent->d_name))
+			    if (rmdir (ent->d_name)) {
 				message (LOG_ERROR,
 					 "Failed to rmdir `%s': %s\n",
 					 ent->d_name, strerror (errno));
+                            }
+                            else {
+                                if (FLAGS_SHOWDEL_P(flags))
+                                    message(LOG_NORMAL, "rmdir %s/%s\n", dirname, ent->d_name);
+                            }
 			}
 		    }
 		}
@@ -532,6 +576,10 @@ cleanupDirectory (const char	   * dirname ,
                                     ent->d_name);
                         }
 #endif
+                        else {
+                            if (FLAGS_SHOWDEL_P(flags))
+                                message(LOG_NORMAL, "rm %s/%s\n", dirname, ent->d_name);
+                        }
 		    }
 		}
 		else {
@@ -628,13 +676,16 @@ main (int     argc,
     struct option const options[] = {
 	{ "help",	no_argument,		NULL, 'h' },
 	{ "test",	no_argument,		NULL, 't' },
-	{ "verbose",	no_argument,		NULL, 'v' },
+	{ "verbose",	optional_argument,	NULL, 'v' },
 	{ "force",	no_argument,		NULL, 'f' },
+	{ "ctime",	no_argument,		NULL, 'c' },
 	{ "mtime",	no_argument,		NULL, 'm' },
 	{ "mtime-dirs",	no_argument,		NULL, 'M' },
 	{ "symlinks",	no_argument,		NULL, 's' },
 	{ "all",	no_argument,		NULL, 'a' },
 	{ "protect",	required_argument,	NULL, 'p' },
+	{ "delay",	optional_argument,	NULL, 'd' },
+	{ "showdeleted",no_argument,	        NULL, 'V' },
 	{ 0, 0, 0, 0 }
     };
 
@@ -644,7 +695,7 @@ main (int     argc,
     while (1) {
 	long_index = 0;
 
-	arg = getopt_long (argc, argv, "htvfmMsa:", options, &long_index);
+	arg = getopt_long(argc, argv, "htvfcmMsa", options, &long_index);
 	if (arg == -1)
 	    break;
 
@@ -656,16 +707,28 @@ main (int     argc,
 
 	case 't':
 	    flags |= FLAGS_TEST;
+            optarg = 0;
 	    /*FALLTHRU*/
 
 	case 'v':
-	    logLevel ? logLevel -= 1 : 0;
+	    if (!optarg)
+                logLevel ? logLevel -= 1 : 0;
+            else if (atoi(optarg) <= 0)
+                logLevel = LOG_NORMAL;
+            else if (atoi(optarg) > 3)
+                logLevel = LOG_REALDEBUG;
+            else
+                logLevel = LOG_NORMAL - atoi(optarg);
 	    break;
 
 	case 'f':
 	    flags |= FLAGS_FORCE;
 	    break;
 
+	case 'c':
+	    flags |= FLAGS_CTIME;
+	    break;
+
 	case 'm':
 	    flags |= FLAGS_MTIME;
 	    break;
@@ -687,6 +750,18 @@ main (int     argc,
 	    protect_argv[k] = NULL;
 	    flags |= FLAGS_PROTECT;
 	    break;
+
+	case 'd':
+	    if (optarg)
+                delayMax = atoi(optarg);
+            else
+                delayMax = 256;
+            message(LOG_DEBUG, "delay is %d seconds max\n", delayMax);
+	    break;
+
+        case 'V':
+            flags |= FLAGS_SHOWDEL;
+            break;
 	}
     }
 
@@ -782,6 +857,32 @@ main (int     argc,
     }
     /* don't need pwd, but leave it allocated for possible debug things below */
 
+    /* Set an alarm, so that we aren't vulnerable to the type of attack that  */
+    /* needs to have multiple concurrent tmpreaper instances running.         */
+    /* A little less than one minute should be enough...                      */
+    /* First, delay a random (small) time to foil wannabee bad guys even more */
+    /* if --delay was given as option.                                        */
+    if (delayMax) {
+        int fd;
+        unsigned char x;
+        unsigned int timetosleep;
+        if ((fd = open("/dev/urandom", O_RDONLY)) >= 0) {
+            read(fd, &x, 1);
+            close(fd);
+            timetosleep = ((x+0.01)/256.0)*delayMax;
+        }
+        else {
+            /* fallback if open /dev/urandom failed */
+            srand(getpid());
+            timetosleep = 1+(int)((delayMax+0.01)*rand()/(RAND_MAX+1.0));
+        }
+	message (LOG_DEBUG, "timetosleep=%u\n", timetosleep);
+        sleep(timetosleep);
+    }
+      
+    signal(SIGALRM, sig_alarmhandler);
+    alarm(55);
+
     /* For each directory in the cleanup list, cd there, glob, then delete */
     for (adir = optind; adir < argc; adir++) {
 	if ((ret = chdir (argv[adir])) != 0) {
